# coding = utf8

import tensorflow as tf
import numpy as np
import faiss


class VectorSearchEngine(object):
    def __init__(self, vectors):
        super().__init__()
        if tf.is_tensor(vectors):
            self.vectors = vectors.numpy()
        else:
            self.vectors = np.array(vectors)
        self.dim = self.vectors.shape[1]

        self.index = faiss.IndexFlatIP(self.dim)
        self.index.add(self.vectors)

    def search(self, query_vectors, k=10):
        query_vectors = np.asarray(query_vectors)
        topK_distances, topK_indices = self.index.search(query_vectors, k)

        return topK_distances, topK_indices


# coding=utf-8


import tensorflow as tf

from grecx.layers.embedding import Embedding


class UserItemEmbedding(tf.keras.Model):

    def __init__(self, num_users, num_items, embedding_size, drop_rate=0.0, global_dropout=False,
                 initializer=None,
                 embeddings_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.user_embedding_layer = Embedding(num_users, embedding_size=embedding_size, drop_rate=drop_rate,
                                              global_dropout=global_dropout,
                                              initializer=initializer, embeddings_regularizer=embeddings_regularizer,
                                              name="user")
        self.item_embedding_layer = Embedding(num_items, embedding_size=embedding_size, drop_rate=drop_rate,
                                              global_dropout=global_dropout,
                                              initializer=initializer, embeddings_regularizer=embeddings_regularizer,
                                              name="item")

    @property
    def user_embeddings(self):
        return self.user_embedding_layer.embeddings

    @property
    def item_embeddings(self):
        return self.item_embedding_layer.embeddings

    def call(self, inputs, training=None, mask=None):
        """
        Embed users and items.

        :param inputs: [nested_user_indices, nested_item_indices]. The output will have the same structure with the input.
            For example:
            - [user_indices, item_indices]
            - [user_indices, [item_indices, neg_item_indices]]

        :param training:
        :param mask:
        :param cache:
        :return:
        """
        user_indices, item_indices = inputs
        embedded_users = self.user_embedding_layer(user_indices, training=training)
        embedded_items = self.item_embedding_layer(item_indices, training=training)
        return embedded_users, embedded_items



# coding=utf-8

import numpy as np
import ctypes
import os

def dcg_score(element_list):
    """
    Discounted Cumulative Gain (DCG)
    The definition of DCG can be found in this paper:
        Azzah Al-Maskari, Mark Sanderson, and Paul Clough. 2007.
        "The relationship between IR effectiveness measures and user satisfaction."

    Parameters:
        element_list - a list of ranks Ex: [5,4,2,2,1]

    Returns:
        score
    """
    score = 0.0
    for order, rel in enumerate(element_list):
        score += (np.power(2.0, rel) - 1.0) / np.log(order + 2)
        # if rel > 0:
        #     print("====")
        #     # score += math.log(2) / math.log((order+2))
        #     score += np.log(2) / np.log(order + 2)
    return score


def ndcg_score(reference, hypothesis):
    """
    Normalized Discounted Cumulative Gain (nDCG)
    Normalized version of DCG:
        nDCG = DCG(hypothesis)/DCG(reference)

    Parameters:
        reference   - a gold standard (perfect) ordering Ex: [5,4,3,2,1]
        hypothesis  - a proposed ordering Ex: [5,2,2,3,1]

    Returns:
        ndcg_score  - normalized score
    """

    return dcg_score(hypothesis)/dcg_score(reference)


def load_c_ndcg_score():

    lib_name = "libranking.dll" if os.name == "nt" else "libranking.so"
    lib_path = os.path.join(os.path.dirname(__file__), lib_name)
    lib = ctypes.CDLL(lib_path)
    c_ndcg_score = lib.ndcg_score
    c_ndcg_score.restype = ctypes.c_double

    def ndcg_score(reference, hypothesis):

        reference = list(reference)
        hypothesis = list(hypothesis)

        num_rels = len(reference)
        rel_type = ctypes.c_int * num_rels

        return c_ndcg_score(
            rel_type(*reference),
            rel_type(*hypothesis),
            ctypes.c_int(num_rels)
        )
    return ndcg_score

try:
    ndcg_score = load_c_ndcg_score()
except:
    print("cannot load c_ndcg_score, use Python version instead")
    pass



def precision_score(reference, hypothesis):
    result = np.sum(hypothesis, dtype=np.float)/len(hypothesis)
    return result


def recall_score(reference, hypothesis):
    result = np.sum(hypothesis, dtype=np.float) / len(reference)
    return result


# coding=utf-8


from grecx.datasets.light_gcn_dataset import LightGCNYelpDataset, LightGCNGowallaDataset, LightGCNAmazonbookDataset
from grecx.datasets.diff_net_dataset import DiffNetYelp, DiffNetFlickr


from . import light_gcn_dataset
from . import diff_net_dataset


# coding=utf-8

import tf_geometric as tfg
import tensorflow as tf
from tf_geometric import SparseAdj
from tf_sparse import sparse_diag_matmul, diag_sparse_matmul
from tf_geometric.utils.graph_utils import convert_edge_to_directed, add_self_loop_edge


class LightGCN(tf.keras.Model):
    CACHE_KEY = "light_gcn_normed_adj"

    def __init__(self, k, edge_drop_rate, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.k = k
        self.edge_drop_rate = edge_drop_rate

    @classmethod
    def build_virtual_edge_index(cls, user_item_edge_index, num_users=None):

        user_index, item_index = user_item_edge_index[0], user_item_edge_index[1]

        if num_users is None:
            num_users = tf.reduce_max(user_index) + 1

        virtual_item_index = item_index + num_users
        virtual_edge_index = tf.stack([user_index, virtual_item_index], axis=0)
        virtual_edge_index, _ = convert_edge_to_directed(virtual_edge_index, merge_modes="sum")
        return virtual_edge_index

    @classmethod
    def norm_adj(cls, edge_index, num_nodes, cache=None):

        if cache is not None:
            cached_data = cache.get(LightGCN.CACHE_KEY, None)
            if cached_data is not None:
                return SparseAdj(cached_data[0], cached_data[1], cached_data[2])
            else:
                if not tf.executing_eagerly():
                    raise Exception("If you want to use cache inside a tf.function, you should manually build the cache before calling the tf.function")

        adj = tfg.SparseAdj(edge_index, shape=[num_nodes, num_nodes])

        deg = adj.segment_sum(axis=-1)
        deg_inv_sqrt = tf.pow(deg, -0.5)
        deg_inv_sqrt = tf.where(
            tf.math.logical_or(tf.math.is_inf(deg_inv_sqrt), tf.math.is_nan(deg_inv_sqrt)),
            tf.zeros_like(deg_inv_sqrt),
            deg_inv_sqrt
        )

        # (D^(-1/2)A)D^(-1/2)
        normed_adj = sparse_diag_matmul(diag_sparse_matmul(deg_inv_sqrt, adj), deg_inv_sqrt)

        if cache is not None:
            cache[LightGCN.CACHE_KEY] = normed_adj.index.numpy(), normed_adj.value.numpy(), normed_adj._shape.numpy()

        return normed_adj

    @classmethod
    def norm_adj_bak(cls, edge_index, num_nodes, cache=None):

        if cache is not None:
            cached_data = cache.get(LightGCN.CACHE_KEY, None)
            if cached_data is not None:
                return cached_data

        adj = tfg.SparseAdj(edge_index, shape=[num_nodes, num_nodes])

        deg = adj.segment_sum(axis=-1)
        deg_inv_sqrt = tf.pow(deg, -0.5)
        deg_inv_sqrt = tf.where(
            tf.math.logical_or(tf.math.is_inf(deg_inv_sqrt), tf.math.is_nan(deg_inv_sqrt)),
            tf.zeros_like(deg_inv_sqrt),
            deg_inv_sqrt
        )

        # (D^(-1/2)A)D^(-1/2)
        normed_adj = sparse_diag_matmul(diag_sparse_matmul(deg_inv_sqrt, adj), deg_inv_sqrt)

        if cache is not None:
            cache[LightGCN.CACHE_KEY] = normed_adj

        return normed_adj

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.
        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        if override:
            graph.cache[LightGCN.CACHE_KEY] = None
        LightGCN.norm_adj(graph.edge_index, graph.num_nodes, cache=graph.cache)

    def call(self, inputs, training=None, mask=None, cache=None):

        x, edge_index = inputs
        num_nodes = tf.shape(x)[0]
        normed_adj = self.norm_adj(edge_index, num_nodes=num_nodes, cache=cache) \
            .dropout(self.edge_drop_rate, training=training)

        h = x
        h_list = [h]

        for _ in range(self.k):
            h = normed_adj @ h
            h_list.append(h)

        h_matrix = tf.stack(h_list, axis=0)
        h = tf.reduce_mean(h_matrix, axis=0)

        return h



# coding=utf-8
import pickle
import os


def save_cache(cache_path, obj):
    with open(cache_path, "wb") as f:
        pickle.dump(obj, f)


def load_cache(cache_path, create_func=None):
    if os.path.exists(cache_path):
        with open(cache_path, "rb") as f:
            return pickle.load(f)
    else:
        if create_func is None:
            raise Exception("cache does not exist: ", cache_path)
        obj = create_func()
        save_cache(cache_path, obj)
        return obj



# coding=utf-8
from itertools import chain

from tqdm import tqdm
import tensorflow as tf
import numpy as np

from grecx.metrics.ranking import ndcg_score, precision_score, recall_score
from grecx.vector_search.vector_search import VectorSearchEngine


# def pred2ndcg_dict(pred_match, num_items, k_list):
#     ndcg_dict = {}
#     for k in k_list:
#         gold = [1] * num_items
#         if len(gold) > k:
#             gold = gold[:k]
#         else:
#             gold = gold + [0] * (k - len(gold))
#
#         ndcg = ndcg_score(gold, pred_match[:k])
#         ndcg_dict["ndcg@{}".format(k)] = ndcg
#     return ndcg_dict
#
#
# def ndcg2mean_ndcg_dict(results, k_list):
#     metrics = ["ndcg@{}".format(K) for K in k_list]
#     mean_ndcg_dict = {}
#     for metric in metrics:
#         scores = [result[metric] for result in results]
#         mean_ndcg = np.mean(scores)
#         mean_ndcg_dict[metric] = mean_ndcg
#     return mean_ndcg_dict


# def evaluate_mean_global_ndcg_score(user_items_dict, user_mask_items_dict, num_items,
#                                     ranking_score_func,
#                                     k_list=[20], user_batch_size=1000, item_batch_size=5000):
#
#     results = []
#     test_users = list(user_items_dict.keys())
#     for batch_user_indices in tqdm(tf.data.Dataset.from_tensor_slices(test_users).batch(user_batch_size)):
#
#         user_rank_score_matrix = []
#
#         for batch_item_indices in tf.data.Dataset.range(num_items).batch(item_batch_size):
#             user_batch_rank_score_matrix = ranking_score_func(batch_user_indices.numpy(), batch_item_indices.numpy())
#             user_rank_score_matrix.append(user_batch_rank_score_matrix)
#
#         user_rank_score_matrix = np.concatenate(user_rank_score_matrix, axis=1)
#
#         for user, user_rank_scores in zip(batch_user_indices, user_rank_score_matrix):
#
#             result = {}
#             results.append(result)
#
#             user = user.numpy()
#             # train_items = train_user_items_dict[user_index]
#             items = user_items_dict[user]
#
#             mask_items = user_mask_items_dict[user]
#
#             # candidate_items = np.array(list(items) + list(user_neg_items_dict[user_index]))
#             pred_items = np.argsort(user_rank_scores)[::-1][:k_list[-1] + len(mask_items)]
#
#             pred_items = [item for item in pred_items if item not in mask_items][:k_list[-1]]
#
#             pred_match = [1.0 if item in items else 0.0 for item in pred_items]
#
#             for k in k_list:
#
#                 gold = [1] * len(items)
#                 if len(gold) > k:
#                     gold = gold[:k]
#                 else:
#                     gold = gold + [0] * (k - len(gold))
#
#                 ndcg = ndcg_score(gold, pred_match[:k])
#                 result["ndcg@{}".format(k)] = ndcg
#
#     metrics = ["ndcg@{}".format(K) for K in k_list]
#     mean_ndcg_dict = {}
#     for metric in metrics:
#         scores = [result[metric] for result in results]
#         mean_ndcg = np.mean(scores)
#         # print(metric, mean_ndcg, len(scores))
#         mean_ndcg_dict[metric] = mean_ndcg
#     return mean_ndcg_dict



# def evaluate_mean_global_all_score(user_items_dict, user_mask_items_dict, num_items,
#                                     ranking_score_func,
#                                     k_list=[20], user_batch_size=1000, item_batch_size=5000):
#
#     results = []
#     test_users = list(user_items_dict.keys())
#     for batch_user_indices in tqdm(tf.data.Dataset.from_tensor_slices(test_users).batch(user_batch_size)):
#
#         user_rank_score_matrix = []
#
#         for batch_item_indices in tf.data.Dataset.range(num_items).batch(item_batch_size):
#             user_batch_rank_score_matrix = ranking_score_func(batch_user_indices.numpy(), batch_item_indices.numpy())
#             user_rank_score_matrix.append(user_batch_rank_score_matrix)
#
#         user_rank_score_matrix = np.concatenate(user_rank_score_matrix, axis=1)
#
#         for user, user_rank_scores in zip(batch_user_indices, user_rank_score_matrix):
#
#             result = {}
#             results.append(result)
#
#             user = user.numpy()
#             # train_items = train_user_items_dict[user_index]
#             items = user_items_dict[user]
#
#             mask_items = user_mask_items_dict[user]
#
#             # candidate_items = np.array(list(items) + list(user_neg_items_dict[user_index]))
#             pred_items = np.argsort(user_rank_scores)[::-1][:k_list[-1] + len(mask_items)]
#
#             pred_items = [item for item in pred_items if item not in mask_items][:k_list[-1]]
#
#             pred_match = [1.0 if item in items else 0.0 for item in pred_items]
#
#             for k in k_list:
#
#                 gold = [1] * len(items)
#                 pr_gold = [1] * len(items)
#                 if len(gold) > k:
#                     gold = gold[:k]
#                 else:
#                     gold = gold + [0] * (k - len(gold))
#
#                 ndcg = ndcg_score(gold, pred_match[:k])
#                 p = precision(pr_gold, pred_match[:k])
#                 r = recall(pr_gold, pred_match[:k])
#                 result["ndcg@{}".format(k)] = ndcg
#                 result["pre@{}".format(k)] = p
#                 result["recall@{}".format(k)] = r
#
#     metrics = [("ndcg@{}".format(K), "pre@{}".format(K), "recall@{}".format(K)) for K in k_list]
#     mean_ndcg_dict = {}
#     mean_pre_dict = {}
#     mean_recall_dict = {}
#     for metric in metrics:
#         ndcg_scores = [result[metric[0]] for result in results]
#         mean_ndcg = np.mean(ndcg_scores)
#         mean_ndcg_dict[metric[0]] = mean_ndcg
#         pre_scores = [result[metric[1]] for result in results]
#         mean_pre = np.mean(pre_scores)
#         mean_pre_dict[metric[1]] = mean_pre
#         recall_scores = [result[metric[2]] for result in results]
#         mean_recall = np.mean(recall_scores)
#         mean_recall_dict[metric[2]] = mean_recall
#     return mean_ndcg_dict, mean_pre_dict, mean_recall_dict

# def evaluate_mean_candidate_ndcg_score(user_items_dict, user_neg_items_dict,
#                                     ranking_score_func,
#                                     k_list=[20], user_batch_size=1000, item_batch_size=5000, num_items=None):
#
#     if num_items is None:
#         num_items = max(max(items) for items in tqdm(chain(user_items_dict.values(), user_neg_items_dict.values())))+1
#         print(num_items)
#
#     results = []
#     test_users = list(user_items_dict.keys())
#     for batch_user_indices in tqdm(tf.data.Dataset.from_tensor_slices(test_users).batch(user_batch_size)):
#
#         user_rank_score_matrix = []
#
#         for batch_item_indices in tf.data.Dataset.range(num_items).batch(item_batch_size):
#             user_batch_rank_score_matrix = ranking_score_func(batch_user_indices.numpy(), batch_item_indices.numpy())
#             user_rank_score_matrix.append(user_batch_rank_score_matrix)
#
#         user_rank_score_matrix = np.concatenate(user_rank_score_matrix, axis=1)
#
#         for user, user_rank_scores in zip(batch_user_indices, user_rank_score_matrix):
#
#             result = {}
#             results.append(result)
#
#             user = user.numpy()
#             # train_items = train_user_items_dict[user_index]
#             items = user_items_dict[user]
#
#             candidate_items = np.array(list(items) + list(user_neg_items_dict[user]))
#             candidate_scores = user_rank_scores[candidate_items]
#             candidate_rank = np.argsort(candidate_scores)[::-1][:k_list[-1]]
#             pred_items = candidate_items[candidate_rank]
#
#             pred_match = [1.0 if item in items else 0.0 for item in pred_items]
#
#             for k in k_list:
#
#                 gold = [1] * len(items)
#                 if len(gold) > k:
#                     gold = gold[:k]
#                 else:
#                     gold = gold + [0] * (k - len(gold))
#
#                 ndcg = ndcg_score(gold, pred_match[:k])
#                 result["ndcg@{}".format(k)] = ndcg
#
#     metrics = ["ndcg@{}".format(K) for K in k_list]
#     mean_ndcg_dict = {}
#     for metric in metrics:
#         scores = [result[metric] for result in results]
#         mean_ndcg = np.mean(scores)
#         # print(metric, mean_ndcg, len(scores))
#         mean_ndcg_dict[metric] = mean_ndcg
#     return mean_ndcg_dict

def score(ground_truth, pred_items, k_list, metrics):
    pred_match = [1 if item in ground_truth else 0 for item in pred_items]

    max_k = k_list[-1]
    if len(ground_truth) > max_k:
        ndcg_gold = [1] * max_k
    else:
        ndcg_gold = [1] * len(ground_truth) + [0] * (max_k - len(ground_truth))

    res_score = []
    for metric in metrics:
        if metric == "ndcg":
            score_func = ndcg_score
        elif metric == "precision":
            score_func = precision_score
        elif metric == "recall":
            score_func = recall_score
        else:
            raise Exception("Not Found Metric : {}".format(metric))

        for k in k_list:
            if metric == "ndcg":
                res_score.append(score_func(ndcg_gold[:k], pred_match[:k]))
            else:
                res_score.append(score_func(ground_truth, pred_match[:k]))

    return res_score


def evaluate_mean_global_metrics(user_items_dict, user_mask_items_dict,
                                 user_embedding, item_embedding,
                                 k_list=[10, 20], metrics=["ndcg"]):

    v_search = VectorSearchEngine(item_embedding)

    if tf.is_tensor(user_embedding):
        user_embedding = user_embedding.numpy()
    else:
        user_embedding = np.asarray(user_embedding)

    user_indices = list(user_items_dict.keys())
    embedded_users = user_embedding[user_indices]
    max_mask_items_length = max(len(user_mask_items_dict[user]) for user in user_indices)

    _, user_rank_pred_items = v_search.search(embedded_users, k_list[-1] + max_mask_items_length)

    res_scores = []
    for user, pred_items in tqdm(zip(user_indices, user_rank_pred_items)):

        items = user_items_dict[user]
        mask_items = user_mask_items_dict[user]
        pred_items = [item for item in pred_items if item not in mask_items][:k_list[-1]]

        res_score = score(items, pred_items, k_list, metrics)

        res_scores.append(res_score)

    res_scores = np.asarray(res_scores)
    names = []
    for metric in metrics:
        for k in k_list:
            names.append("{}@{}".format(metric, k))

    # return list(zip(names, np.mean(res_scores, axis=0, keepdims=False)))
    return dict(zip(names, np.mean(res_scores, axis=0, keepdims=False)))


# def evaluate_mean_global_metrics(user_items_dict, user_mask_items_dict,
#                                  user_embedding, item_embedding,
#                                  k_list=[10, 20], metrics=["ndcg"]):
#
#     v_search = VectorSearch(item_embedding)
#
#     if isinstance(user_embedding, tf.Tensor):
#         user_embedding = np.asarray(user_embedding)
#
#     user_indices = list(user_items_dict.keys())
#     embedded_users = user_embedding[user_indices]
#     max_mask_items_length = max(len(user_mask_items_dict[user]) for user in user_indices)
#
#     _, user_rank_pred_items = v_search.search(embedded_users, k_list[-1] + max_mask_items_length)
#
#     results = []
#     for user, pred_items in tqdm(zip(user_indices, user_rank_pred_items)):
#
#         items = user_items_dict[user]
#         mask_items = user_mask_items_dict[user]
#         pred_items = [item for item in pred_items if item not in mask_items][:k_list[-1]]
#
#         pred_match = [1.0 if item in items else 0.0 for item in pred_items]
#
#         results.append(pred2ndcg_dict(pred_match, len(items), k_list))
#
#     return ndcg2mean_ndcg_dict(results, k_list)





# coding=utf-8

from tf_geometric.data.dataset import DownloadableDataset
import os
import numpy as np


class LightGCNDataset(DownloadableDataset):
    """
    Datasets used in the LightGCN model: https://github.com/maenzhier/grecx_datasets
    """

    def __init__(self, dataset_name, dataset_root_path=None):
        """

        :param dataset_name: "light_gcn_yelp" | "light_gcn_gowalla" | "light_gcn_amazon-book"
        :param featureless:
        :param download_urls:
        :param dataset_root_path:
        """
        super().__init__(dataset_name,
                         download_urls=[
                             "https://github.com/maenzhier/grecx_datasets/raw/main/{}/{}.zip".format(
                                 dataset_name.replace("light_gcn_", ""), dataset_name)
                         ],
                         download_file_name="{}.zip".format(dataset_name),
                         cache_name="cache.p",
                         dataset_root_path=dataset_root_path)

    def _read_edge_info(self, file_path):
        edge_dict = {}
        edges = []

        with open(file_path, "r", encoding="utf-8") as f:
            for l in f.readlines():
                if len(l) > 0:
                    try:
                        l = l.strip('\n').split(' ')
                        items = []
                        uid = int(l[0])
                        for i in l[1:]:
                            i = int(i)
                            items.append(i)
                            edges.append([uid, i])
                        if uid not in edge_dict:
                            edge_dict[uid] = set(items)
                        else:
                            item_set = edge_dict[uid]
                            edge_dict[uid] = set(items).union(item_set)
                    except Exception:
                        continue

        edges = np.array(edges)
        return edge_dict, edges

    def process(self):

        name = self.dataset_name.replace("light_gcn_", "")
        path = os.path.join(self.raw_root_path, "light_gcn_{}".format(name))

        # data = _LightGCNData(path=data_dir)

        train_file = path + '/train.txt'
        test_file = path + '/test.txt'

        train_user_items_dict, train_user_item_edges = self._read_edge_info(train_file)
        test_user_items_dict, test_user_item_edges = self._read_edge_info(test_file)

        user_item_edges = np.concatenate([train_user_item_edges, test_user_item_edges], axis=0)
        index = np.arange(user_item_edges.shape[0])
        num_train_edges = train_user_item_edges.shape[0]
        train_index, test_index = index[:num_train_edges], index[num_train_edges:]
        num_users, num_items = user_item_edges.max(axis=0) + 1

        return {
            "num_users": num_users,
            "num_items": num_items,
            "user_item_edges": user_item_edges,
            "train_index": train_index,
            "test_index": test_index,
            "train_user_items_dict": train_user_items_dict,
            "test_user_items_dict": test_user_items_dict,
        }


class LightGCNYelpDataset(LightGCNDataset):
    """
    The Yelp dataset used in the LightGCN model: https://github.com/maenzhier/grecx_datasets/tree/main/yelp
    """

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_name="light_gcn_yelp", dataset_root_path=dataset_root_path)


class LightGCNGowallaDataset(LightGCNDataset):
    """
    The Gowalla dataset used in the LightGCN model: https://github.com/maenzhier/grecx_datasets/tree/main/gowalla
    """

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_name="light_gcn_gowalla", dataset_root_path=dataset_root_path)


class LightGCNAmazonbookDataset(LightGCNDataset):
    """
    The Amazonbook dataset used in the LightGCN model: https://github.com/maenzhier/grecx_datasets/tree/main/amazon-book
    """

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_name="light_gcn_amazon-book", dataset_root_path=dataset_root_path)



# coding=utf-8
# import os
# os.environ["CUDA_VISIBLE_DEVICES"] = "5"

# replace data_base_dir with path of your own data directory
# data_base_dir = os.path.join(os.path.dirname(__file__), "../data")
# dataset = "yelp"
# diff_dataset = "flickr"
# data_dir = os.path.join(data_base_dir, dataset)


# train_path = os.path.join(data_dir, "{}.train.rating".format(dataset))
# test_path = os.path.join(data_dir, "{}.test.rating".format(dataset))
# links_path = os.path.join(data_dir, "{}.links".format(dataset))
# cached_data_path = os.path.join(data_dir, "cached.p")
# cached_neg_data_path = os.path.join(data_dir, "cached_neg.p")

num_negs = 1000
# test_neg_path = os.path.join(data_dir, "{}.test.negative".format(diff_dataset))

embedding_size = 64
# embedding_size = 256

# print(os.listdir(data_dir))



import setuptools
import io
from setuptools import setup, find_packages

setup(
    name="grecx",
    python_requires='>3.5.0',
    version="0.0.4.9",
    author="Desheng Cai",
    author_email="caidsml@gmail.com",
    packages=find_packages(
        exclude=[
            'benchmarks',
            'data',
            'demo',
            'dist',
            'doc',
            'docs',
            'logs',
            'models',
            'test'
        ]
    ),
    install_requires=[
        "tf_geometric >= 0.0.77",
        "numpy >= 1.17.4",
        # "tensorflow == 2.4.1",
        "scikit-learn >= 0.22",
        "tqdm",
        # "Sphinx == 3.5.4",
        "faiss-cpu"
    ],
    extras_require={

    },
    package_data={'grecx': ["metrics/libranking.so", "metrics/libranking.dll"]},
    description="An Efficient and Unified Benchmark for GNN-based Recommendation.",
    license="GNU General Public License v3.0 (See LICENSE)",
    long_description=io.open("README.rst", mode="r", encoding="utf-8").read(),
    # long_description="An Efficient and Unified Benchmark for GNN-based Recommendation.",
    long_description_content_type='text/x-rst',
    url="https://github.com/maenzhier/GRecX"
)


# coding=utf-8

import tensorflow as tf
import tf_sparse as tfs
from grecx.layers.light_gcn import LightGCN


class NGCFConv(tf.keras.Model):
    """
    Each NGCF Convolutional Layer
    """

    def __init__(self, dense_activation=tf.nn.leaky_relu, edge_drop_rate=0.0, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.dense_activation = dense_activation
        self.gcn_dense = None
        self.interaction_dense = None
        self.edge_drop_rate = edge_drop_rate

    def build(self, input_shape):
        x_shape, _ = input_shape

        self.gcn_dense = tf.keras.layers.Dense(x_shape[1], activation=self.dense_activation)
        self.interaction_dense = tf.keras.layers.Dense(x_shape[1], activation=self.dense_activation)

    @classmethod
    def build_virtual_edge_index(cls, user_item_edge_index, num_users=None):
        return LightGCN.build_virtual_edge_index(user_item_edge_index, num_users)

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.
        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        if override:
            graph.cache[LightGCN.CACHE_KEY] = None
        LightGCN.norm_adj(graph.edge_index, graph.num_nodes, cache=graph.cache)

    def call(self, inputs, training=None, mask=None, cache=None):
        x, edge_index = inputs[0], inputs[1]

        num_nodes = tfs.shape(x)[0]
        normed_adj = LightGCN.norm_adj(edge_index, num_nodes=num_nodes, cache=cache) \
            .dropout(self.edge_drop_rate, training=training)

        h = normed_adj @ x
        h = self.gcn_dense(h) + self.interaction_dense(x * h)

        return h


class NGCF(tf.keras.Model):

    def __init__(self, k=3, dense_activation=tf.nn.leaky_relu, edge_drop_rate=0.0, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ngcf_convs = [NGCFConv(dense_activation=dense_activation,edge_drop_rate=edge_drop_rate) for _ in range(k)]

        for i, ngcf_conv in enumerate(self.ngcf_convs):
            setattr(self, "ngcf_conv{}".format(i), ngcf_conv)


    @classmethod
    def build_virtual_edge_index(cls, user_item_edge_index, num_users=None):
        return LightGCN.build_virtual_edge_index(user_item_edge_index, num_users)

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.
        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        self.ngcf_convs[0].build_cache_for_graph(graph, override=override)

    def call(self, inputs, training=None, mask=None, cache=None):
        x, edge_index = inputs

        h = x
        h_list = []

        for ngcf_conv in self.ngcf_convs:
            h = ngcf_conv([h, edge_index], training=training, cache=cache)
            h_list.append(h)

        h_list = [tf.nn.l2_normalize(h, axis=-1) for h in h_list]

        h = tf.concat([x] + h_list, axis=-1)

        return h



# coding=utf-8
from multiprocessing import Pool

from tf_geometric.data.graph import Graph
from tf_geometric.data.dataset import DownloadableDataset
import os
import numpy as np
import scipy.sparse as sp
from tqdm import tqdm


class DiffNetDataset(DownloadableDataset):

    """
    Datasets used in the DiffNet model: https://github.com/maenzhier/grecx_datasets
    """

    def __init__(self, dataset_name, featureless=True, download_urls=None, dataset_root_path=None):
        """

        :param dataset_name: "diff_net_yelp" | "diff_net_flickr"
        :param featureless:
        :param download_urls:
        :param dataset_root_path:
        """
        super().__init__(dataset_name, download_urls,
                         download_file_name="{}.zip".format(dataset_name),
                         cache_name="cache.p",
                         dataset_root_path=dataset_root_path)
        self.featureless = featureless

    def read_edge_info(self, file_path):
        edge_dict = {}
        edges = []

        with open(file_path, "r", encoding="utf-8") as f:
            for line in tqdm(f):

                items = [int(item) for item in line.split()[:3]]
                a = items[0]
                b = items[1]
                rating = items[2]

                if a not in edge_dict:
                    item_set = set()
                    edge_dict[a] = item_set
                else:
                    item_set = edge_dict[a]

                if rating > 0:
                    item_set.add(b)
                    edges.append([a, b])

        edges = np.array(edges)
        return edge_dict, edges


    def process(self):
        if "Yelp" in self.dataset_name:
            name = "yelp"
        else:
            name = "flickr"
        data_dir = os.path.join(self.raw_root_path, "diffnet_{}".format(name))
        print(data_dir)

        train_path = os.path.join(data_dir, "{}.train.rating".format(name))
        test_path = os.path.join(data_dir, "{}.test.rating".format(name))
        links_path = os.path.join(data_dir, "{}.links".format(name))

        train_user_items_dict, train_user_item_edges = self.read_edge_info(train_path)
        test_user_items_dict, test_user_item_edges = self.read_edge_info(test_path)
        user_users_dict, user_user_edges = self.read_edge_info(links_path)

        user_item_edges = np.concatenate([train_user_item_edges, test_user_item_edges], axis=0)
        num_users, num_items = user_item_edges.max(axis=0) + 1

        num_negs = 1000

        def build_user_neg_items_dict():
            user_neg_items_dict = {}
            all_items = list(range(num_items))
            for user in tqdm(test_user_items_dict):
                p = np.ones([num_items])
                p[list(train_user_items_dict[user])] = 0.0
                p[list(test_user_items_dict[user])] = 0.0
                p /= p.sum()
                neg_items = np.random.choice(all_items, num_negs, replace=False, p=p)
                user_neg_items_dict[user] = neg_items
            return user_neg_items_dict

        user_neg_items_dict = build_user_neg_items_dict()

        return num_users, num_items, train_user_item_edges, test_user_items_dict, user_user_edges, user_neg_items_dict


class DiffNetYelp(DiffNetDataset):

    """
    The Yelp dataset used in the DiffNet model: https://github.com/maenzhier/grecx_datasets/tree/main/yelp
    """

    def __init__(self, dataset_root_path=None, featureless=True):
        super().__init__(
            dataset_name="DiffNetYelp",
            featureless=featureless,
            download_urls="https://github.com/maenzhier/grecx_datasets/raw/main/yelp/diff_net_yelp.zip",
            dataset_root_path=dataset_root_path,
        )


class DiffNetFlickr(DiffNetDataset):
    """
    The Flickr dataset used in the DiffNet model: https://github.com/maenzhier/grecx_datasets/tree/main/yelp
    """
    def __init__(self, dataset_root_path=None, featureless=True):
        super().__init__(
            dataset_name="diff_net_flickr",
            featureless=featureless,
            download_urls="https://github.com/maenzhier/grecx_datasets/raw/main/flickr/diff_net_flickr.zip",
            dataset_root_path=dataset_root_path,
        )


# coding=utf-8


import tensorflow as tf
import numpy as np


class Embedding(tf.keras.Model):

    def __init__(self, num_embeddings, embedding_size, drop_rate=0.0, global_dropout=False,
                 initializer=None,
                 embeddings_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.global_dropout = global_dropout

        if initializer is None:
            initializer = tf.keras.initializers.Constant(value=tf.random.truncated_normal([num_embeddings, embedding_size], stddev=1.0/np.sqrt(embedding_size)))

        self.embeddings = self.add_weight("embeddings", shape=[num_embeddings, embedding_size], initializer=initializer, regularizer=embeddings_regularizer)
        if drop_rate is not None and drop_rate > 0.0:
            self.dropout = tf.keras.layers.Dropout(drop_rate)
        else:
            self.dropout = None

    def call(self, inputs, training=None, mask=None, cache=None):


        embeddings = self.embeddings

        if self.global_dropout and self.dropout is not None:
            embeddings = self.dropout(embeddings, training=training)

        def gather_func(indices):
            return tf.gather(embeddings, indices)

        output = tf.nest.map_structure(gather_func, inputs)

        if not self.global_dropout and self.dropout is not None:
            def dropout_func(embedded):
                return self.dropout(embedded, training=training)
            output = tf.nest.map_structure(dropout_func, output)

        return output

    # def call(self, inputs, training=None, mask=None, cache=None):
    #
    #     inputs_is_list = isinstance(inputs, list)
    #
    #     embeddings = self.embeddings
    #
    #     if self.global_dropout and self.dropout is not None:
    #         embeddings = self.dropout(embeddings, training=training)
    #
    #
    #
    #     if inputs_is_list:
    #         embedded_list = [tf.gather(embeddings, indices) for indices in inputs]
    #     else:
    #         embedded = tf.gather(embeddings, inputs)
    #
    #     if not self.global_dropout and self.dropout is not None:
    #         if inputs_is_list:
    #             output = [self.dropout(embedded, training=training) for embedded in embedded_list]
    #         else:
    #             output = self.dropout(embedded, training=training)
    #
    #     return output
    #
